@@ -2,7 +2,9 @@
 
 import copy
 import logging
+import time
 
+from pkg_resources import parse_version
 from typing import Any
 from typing import Callable
 from typing import Dict
@@ -13,24 +15,15 @@ from typing import Union
 import lightgbm as lgb
 import numpy as np
 import optuna
+import sklearn
 
 from sklearn.base import ClassifierMixin
 from sklearn.base import RegressorMixin
 from sklearn.model_selection import BaseCrossValidator
 from sklearn.preprocessing import LabelEncoder
 from sklearn.utils import check_random_state
-from sklearn.utils import safe_indexing
 from sklearn.utils.validation import check_is_fitted
 
-try:  # lightgbm<=2.2.3
-    from lightgbm.sklearn import _eval_function_wrapper as _EvalFunctionWrapper
-    from lightgbm.sklearn import (
-        _objective_function_wrapper as _ObjectiveFunctionWrapper,
-    )
-except ImportError:
-    from lightgbm.sklearn import _EvalFunctionWrapper
-    from lightgbm.sklearn import _ObjectiveFunctionWrapper
-
 from .utils import check_cv
 from .utils import check_fit_params
 from .utils import check_X
@@ -39,6 +32,20 @@ from .utils import ONE_DIM_ARRAYLIKE_TYPE
 from .utils import RANDOM_STATE_TYPE
 from .utils import TWO_DIM_ARRAYLIKE_TYPE
 
+if parse_version(lgb.__version__) >= parse_version("2.3"):
+    from lightgbm.sklearn import _EvalFunctionWrapper
+    from lightgbm.sklearn import _ObjectiveFunctionWrapper
+else:
+    from lightgbm.sklearn import _eval_function_wrapper as _EvalFunctionWrapper
+    from lightgbm.sklearn import (
+        _objective_function_wrapper as _ObjectiveFunctionWrapper,
+    )
+
+if parse_version(sklearn.__version__) >= parse_version("0.22"):
+    from sklearn.utils import _safe_indexing as safe_indexing
+else:
+    from sklearn.utils import safe_indexing
+
 MAX_INT = np.iinfo(np.int32).max
 
 OBJECTIVE2METRIC = {
@@ -71,26 +78,6 @@ OBJECTIVE2METRIC = {
     "tweedie": "tweedie",
 }
 
-DEFAULT_PARAM_DISTRIBUTIONS = {
-    # "boosting_type": optuna.distributions.CategoricalDistribution(
-    #     ["gbdt", "rf"]
-    # ),
-    "colsample_bytree": optuna.distributions.DiscreteUniformDistribution(
-        0.1, 1.0, 0.05
-    ),
-    "min_child_samples": optuna.distributions.IntUniformDistribution(1, 100),
-    # "min_child_weight": optuna.distributions.LogUniformDistribution(
-    #     1e-03, 10.0
-    # ),
-    "num_leaves": optuna.distributions.IntUniformDistribution(2, 127),
-    "reg_alpha": optuna.distributions.LogUniformDistribution(1e-09, 10.0),
-    "reg_lambda": optuna.distributions.LogUniformDistribution(1e-09, 10.0),
-    "subsample": optuna.distributions.DiscreteUniformDistribution(
-        0.5, 0.95, 0.05
-    ),
-    "subsample_freq": optuna.distributions.IntUniformDistribution(1, 10),
-}
-
 
 def _is_higher_better(metric: str) -> bool:
     return metric in ["auc"]
@@ -98,8 +85,8 @@ def _is_higher_better(metric: str) -> bool:
 
 class _LightGBMExtractionCallback(object):
     def __init__(self) -> None:
-        self._best_iteration: Optional[int] = None
-        self._boosters: Optional[List[lgb.Booster]] = None
+        self._best_iteration = None  # type: Optional[int]
+        self._boosters = None  # type: Optional[List[lgb.Booster]]
 
     def __call__(self, env: LightGBMCallbackEnv) -> None:
         self._best_iteration = env.iteration + 1
@@ -107,21 +94,13 @@ class _LightGBMExtractionCallback(object):
 
 
 class _Objective(object):
-    @property
-    def _param_distributions(
-        self,
-    ) -> Dict[str, optuna.distributions.BaseDistribution]:
-        if self.param_distributions is None:
-            return DEFAULT_PARAM_DISTRIBUTIONS
-
-        return self.param_distributions
-
     def __init__(
         self,
         params: Dict[str, Any],
         dataset: lgb.Dataset,
         eval_name: str,
         is_higher_better: bool,
+        n_samples: int,
         callbacks: Optional[List[Callable]] = None,
         categorical_feature: Union[List[int], List[str], str] = "auto",
         cv: Optional[BaseCrossValidator] = None,
@@ -147,14 +126,15 @@ class _Objective(object):
         self.fobj = fobj
         self.is_higher_better = is_higher_better
         self.n_estimators = n_estimators
+        self.n_samples = n_samples
         self.params = params
         self.param_distributions = param_distributions
 
     def __call__(self, trial: optuna.trial.Trial) -> float:
-        params: Dict[str, Any] = self._get_params(trial)
+        params = self._get_params(trial)  # type: Dict[str, Any]
         dataset = copy.copy(self.dataset)
-        callbacks: List[Callable] = self._get_callbacks(trial)
-        eval_hist: Dict[str, List[float]] = lgb.cv(
+        callbacks = self._get_callbacks(trial)  # type: List[Callable]
+        eval_hist = lgb.cv(
             params,
             dataset,
             callbacks=callbacks,
@@ -165,9 +145,9 @@ class _Objective(object):
             fobj=self.fobj,
             folds=self.cv,
             num_boost_round=self.n_estimators,
-        )
-        value: float = eval_hist[f"{self.eval_name}-mean"][-1]
-        is_best_trial: bool = True
+        )  # Dict[str, List[float]]
+        value = eval_hist["{}-mean".format(self.eval_name)][-1]  # type: float
+        is_best_trial = True  # type: bool
 
         try:
             is_best_trial = (
@@ -177,11 +157,11 @@ class _Objective(object):
             pass
 
         if is_best_trial:
-            best_iteration: int = callbacks[0]._best_iteration  # type: ignore
-            boosters: List[lgb.Booster] = (
+            best_iteration = callbacks[0]._best_iteration  # type: ignore
+            boosters = (
                 callbacks[0]._boosters  # type: ignore
-            )
-            representations: List[str] = []
+            )  # type: List[lgb.Booster]
+            representations = []  # type: List[str]
 
             for b in boosters:
                 b.free_dataset()
@@ -193,17 +173,17 @@ class _Objective(object):
         return value
 
     def _get_callbacks(self, trial: optuna.trial.Trial) -> List[Callable]:
-        extraction_callback: _LightGBMExtractionCallback = (
+        extraction_callback = (
             _LightGBMExtractionCallback()
-        )
-        callbacks: List[Callable] = [extraction_callback]
+        )  # type: _LightGBMExtractionCallback
+        callbacks = [extraction_callback]  # type: List[Callable]
 
         if self.enable_pruning:
-            pruning_callback: optuna.integration.LightGBMPruningCallback = (
+            pruning_callback = (
                 optuna.integration.LightGBMPruningCallback(
                     trial, self.eval_name
                 )
-            )
+            )  # type: optuna.integration.LightGBMPruningCallback
 
             callbacks.append(pruning_callback)
 
@@ -213,9 +193,39 @@ class _Objective(object):
         return callbacks
 
     def _get_params(self, trial: optuna.trial.Trial) -> Dict[str, Any]:
-        params: Dict[str, Any] = self.params.copy()
+        params = self.params.copy()  # type: Dict[str, Any]
 
-        for name, distribution in self._param_distributions.items():
+        if self.param_distributions is None:
+            params["colsample_bytree"] = trial.suggest_discrete_uniform(
+                "colsample_bytree", 0.1, 1.0, 0.05
+            )
+            params["max_depth"] = trial.suggest_int("max_depth", 1, 7)
+            params["num_leaves"] = trial.suggest_int(
+                "num_leaves", 2, 2 ** params["max_depth"]
+            )
+            params["min_child_samples"] = trial.suggest_int(
+                "min_child_samples",
+                1,
+                max(1, int(self.n_samples / params["num_leaves"])),
+            )
+            params["reg_alpha"] = trial.suggest_loguniform(
+                "reg_alpha", 1e-09, 10.0
+            )
+            params["reg_lambda"] = trial.suggest_loguniform(
+                "reg_lambda", 1e-09, 10.0
+            )
+
+            if params["boosting_type"] != "goss":
+                params["subsample"] = trial.suggest_discrete_uniform(
+                    "subsample", 0.5, 0.95, 0.05
+                )
+                params["subsample_freq"] = trial.suggest_int(
+                    "subsample_freq", 1, 10
+                )
+
+            return params
+
+        for name, distribution in self.param_distributions.items():
             params[name] = trial._suggest(name, distribution)
 
         return params
@@ -236,16 +246,16 @@ class _VotingBooster(object):
     def from_representations(
         cls, representations: List[str], weights: Optional[np.ndarray] = None
     ) -> "_VotingBooster":
-        try:  # lightgbm<=2.2.3
-            boosters = [
-                lgb.Booster(params={"model_str": model_str})
-                for model_str in representations
-            ]
-        except TypeError:
+        if parse_version(lgb.__version__) >= parse_version("2.3"):
             boosters = [
                 lgb.Booster(model_str=model_str, silent=True)
                 for model_str in representations
             ]
+        else:
+            boosters = [
+                lgb.Booster(params={"model_str": model_str})
+                for model_str in representations
+            ]
 
         return cls(boosters, weights=weights)
 
@@ -274,8 +284,8 @@ class _BaseOGBMModel(lgb.LGBMModel):
         num_leaves: int = 31,
         max_depth: int = -1,
         learning_rate: float = 0.1,
-        n_estimators: int = 1_000,
-        subsample_for_bin: int = 200_000,
+        n_estimators: int = 1000,
+        subsample_for_bin: int = 200000,
         objective: Optional[Union[Callable, str]] = None,
         class_weight: Optional[Union[Dict[str, float], str]] = None,
         min_split_gain: float = 0.0,
@@ -292,13 +302,13 @@ class _BaseOGBMModel(lgb.LGBMModel):
         importance_type: str = "split",
         cv: Union[BaseCrossValidator, int] = 5,
         enable_pruning: bool = False,
-        n_trials: int = 10,
+        n_trials: int = 20,
         param_distributions: Optional[
             Dict[str, optuna.distributions.BaseDistribution]
         ] = None,
         study: Optional[optuna.study.Study] = None,
         timeout: Optional[float] = None,
-        **kwargs: Any,
+        **kwargs: Any
     ) -> None:
         super().__init__(
             boosting_type=boosting_type,
@@ -320,7 +330,7 @@ class _BaseOGBMModel(lgb.LGBMModel):
             subsample=subsample,
             subsample_for_bin=subsample_for_bin,
             subsample_freq=subsample_freq,
-            **kwargs,
+            **kwargs
         )
 
         self.cv = cv
@@ -406,12 +416,14 @@ class _BaseOGBMModel(lgb.LGBMModel):
         X: TWO_DIM_ARRAYLIKE_TYPE,
         y: ONE_DIM_ARRAYLIKE_TYPE,
         sample_weight: Optional[ONE_DIM_ARRAYLIKE_TYPE] = None,
-        groups: Optional[ONE_DIM_ARRAYLIKE_TYPE] = None,
-        callbacks: Optional[List[Callable]] = None,
-        categorical_feature: Union[List[int], List[str], str] = "auto",
-        early_stopping_rounds: Optional[int] = 10,
+        group: Optional[ONE_DIM_ARRAYLIKE_TYPE] = None,
         eval_metric: Optional[Union[Callable, str]] = None,
+        early_stopping_rounds: Optional[int] = 10,
         feature_name: Union[List[str], str] = "auto",
+        categorical_feature: Union[List[int], List[str], str] = "auto",
+        callbacks: Optional[List[Callable]] = None,
+        groups: Optional[ONE_DIM_ARRAYLIKE_TYPE] = None,
+        **fit_params: Any
     ) -> "_BaseOGBMModel":
         """Fit the model according to the given training data.
 
@@ -426,25 +438,31 @@ class _BaseOGBMModel(lgb.LGBMModel):
         sample_weight
             Weights of training data.
 
-        groups
-            Group labels for the samples used while splitting the dataset into
-            train/test set.
-
-        callbacks
-            List of callback functions that are applied at each iteration.
-
-        categorical_feature
-            Categorical features.
-
-        early_stopping_rounds
-            Used to activate early stopping.
+        group
+            Group data of training data.
 
         eval_metric
             Evaluation metric.
 
+        early_stopping_rounds
+            Used to activate early stopping.
+
         feature_name
             Feature names.
 
+        categorical_feature
+            Categorical features.
+
+        callbacks
+            List of callback functions that are applied at each iteration.
+
+        groups
+            Group labels for the samples used while splitting the dataset into
+            train/test set. If `group` is not None, this parameter is ignored.
+
+        **fit_params
+            Always ignored, exists for compatibility.
+
         Returns
         -------
         self
@@ -460,7 +478,7 @@ class _BaseOGBMModel(lgb.LGBMModel):
             force_all_finite=False,
         )
 
-        _, self._n_features = X.shape
+        n_samples, self._n_features = X.shape
 
         is_classifier = self._estimator_type == "classifier"
         cv = check_cv(self.cv, y, is_classifier)
@@ -469,15 +487,19 @@ class _BaseOGBMModel(lgb.LGBMModel):
 
         params = self.get_params()
 
-        params.pop("class_weight", None)
-        params.pop("cv")
-        params.pop("enable_pruning")
-        params.pop("importance_type")
-        params.pop("n_estimators")
-        params.pop("n_trials")
-        params.pop("param_distributions")
-        params.pop("study")
-        params.pop("timeout")
+        for attr in (
+            "class_weight",
+            "cv",
+            "enable_pruning",
+            "importance_type",
+            "n_estimators",
+            "n_trials",
+            "param_distributions",
+            "refit",
+            "study",
+            "timeout",
+        ):
+            params.pop(attr, None)
 
         params["random_state"] = seed
         params["verbose"] = -1
@@ -528,9 +550,8 @@ class _BaseOGBMModel(lgb.LGBMModel):
         else:
             self.study_ = self.study
 
-        if groups is None:
-            group = None
-        else:
+        # https://github.com/microsoft/LightGBM/issues/2319
+        if group is None and groups is not None:
             indices = np.argsort(groups)
             X = safe_indexing(X, indices)
             y = safe_indexing(y, indices)
@@ -545,6 +566,7 @@ class _BaseOGBMModel(lgb.LGBMModel):
             dataset,
             eval_name,
             is_higher_better,
+            n_samples,
             callbacks=callbacks,
             categorical_feature=categorical_feature,
             cv=cv,
@@ -557,28 +579,28 @@ class _BaseOGBMModel(lgb.LGBMModel):
             param_distributions=self.param_distributions,
         )
 
+        logger = logging.getLogger(__name__)
+
+        logger.info("Searching the best hyperparameters...")
+
         self.study_.optimize(
             objective, catch=(), n_trials=self.n_trials, timeout=self.timeout
         )
 
+        logger.info("Finished hyperparemeter search!")
+
         self.best_params_ = {**params, **self.study_.best_params}
         self._best_iteration = self.study_.user_attrs["best_iteration"]
+        self._best_score = self.study_.best_value
         self.n_splits_ = cv.get_n_splits(X, y, groups=groups)
 
-        logger = logging.getLogger(__name__)
-
-        logger.info(f"The best_iteration is {self._best_iteration}.")
-
-        weights = np.array(
-            [
-                np.sum(sample_weight[train])
-                for train, _ in cv.split(X, y, groups=groups)
-            ]
-        )
-
-        self._Booster: Union[lgb.Booster, _VotingBooster]
+        logger.info("The best_iteration is {}.".format(self._best_iteration))
 
         if self.refit:
+            logger.info("Refitting the estimator...")
+
+            start_time = time.perf_counter()
+
             self._Booster = self._train_booster(
                 X,
                 y,
@@ -587,7 +609,21 @@ class _BaseOGBMModel(lgb.LGBMModel):
                 categorical_feature=categorical_feature,
                 feature_name=feature_name,
             )
+            self.refit_time_ = time.perf_counter() - start_time
+
+            logger.info(
+                "Finished refitting! "
+                "(elapsed time: {:.3f} sec.)".format(self.refit_time_)
+            )
+
         else:
+            weights = np.array(
+                [
+                    np.sum(sample_weight[train])
+                    for train, _ in cv.split(X, y, groups=groups)
+                ]
+            )
+
             self._Booster = _VotingBooster.from_representations(
                 self.study_.user_attrs["representations"], weights=weights
             )
@@ -691,6 +727,9 @@ class OGBMClassifier(_BaseOGBMModel, ClassifierMixin):
     best_params_
         Parameters of the best trial in the `Study`.
 
+    best_score_
+        Mean cross-validated score of the best estimator.
+
     booster_
         Trained booster.
 
@@ -706,6 +745,9 @@ class OGBMClassifier(_BaseOGBMModel, ClassifierMixin):
     study_
         Actual study.
 
+    refit_time_
+        Time for refitting the best estimator.
+
     Examples
     --------
     >>> from optgbm.sklearn import OGBMClassifier
@@ -714,8 +756,7 @@ class OGBMClassifier(_BaseOGBMModel, ClassifierMixin):
     >>> X, y = load_iris(return_X_y=True)
     >>> clf.fit(X, y)
     OGBMClassifier(...)
-    >>> clf.score(X, y)
-    0.9...
+    >>> y_pred = clf.predict(X)
     """
 
     @property
@@ -728,9 +769,16 @@ class OGBMClassifier(_BaseOGBMModel, ClassifierMixin):
     @property
     def n_classes_(self) -> int:
         """Get the number of classes."""
+        self._check_is_fitted()
+
         return self._n_classes
 
-    def predict(self, X: TWO_DIM_ARRAYLIKE_TYPE) -> ONE_DIM_ARRAYLIKE_TYPE:
+    def predict(
+        self,
+        X: TWO_DIM_ARRAYLIKE_TYPE,
+        num_iteration: Optional[int] = None,
+        **predict_params: Any
+    ) -> ONE_DIM_ARRAYLIKE_TYPE:
         """Predict using the fitted model.
 
         Parameters
@@ -738,18 +786,31 @@ class OGBMClassifier(_BaseOGBMModel, ClassifierMixin):
         X
             Data.
 
+        num_iteration
+            Limit number of iterations in the prediction.
+
+        **predict_params
+            Always ignored, exists for compatibility.
+
         Returns
         -------
         y_pred
             Predicted values.
         """
-        probas = self.predict_proba(X)
+        probas = self.predict_proba(
+            X,
+            num_iteration=num_iteration,
+            **predict_params
+        )
         class_index = np.argmax(probas, axis=1)
 
         return self.encoder_.inverse_transform(class_index)
 
     def predict_proba(
-        self, X: TWO_DIM_ARRAYLIKE_TYPE
+        self,
+        X: TWO_DIM_ARRAYLIKE_TYPE,
+        num_iteration: Optional[int] = None,
+        **predict_params: Any
     ) -> TWO_DIM_ARRAYLIKE_TYPE:
         """Predict class probabilities for data.
 
@@ -758,6 +819,12 @@ class OGBMClassifier(_BaseOGBMModel, ClassifierMixin):
         X
             Data.
 
+        num_iteration
+            Limit number of iterations in the prediction.
+
+        **predict_params
+            Always ignored, exists for compatibility.
+
         Returns
         -------
         p
@@ -768,7 +835,7 @@ class OGBMClassifier(_BaseOGBMModel, ClassifierMixin):
         X = check_X(
             X, accept_sparse=True, estimator=self, force_all_finite=False
         )
-        preds = self._Booster.predict(X)
+        preds = self._Booster.predict(X, num_iteration=num_iteration)
 
         if self._n_classes > 2:
             return preds
@@ -872,6 +939,9 @@ class OGBMRegressor(_BaseOGBMModel, RegressorMixin):
     best_params_
         Parameters of the best trial in the `Study`.
 
+    best_score_
+        Mean cross-validated score of the best estimator.
+
     booster_
         Trained booster.
 
@@ -884,6 +954,9 @@ class OGBMRegressor(_BaseOGBMModel, RegressorMixin):
     study_
         Actual study.
 
+    refit_time_
+        Time for refitting the best estimator.
+
     Examples
     --------
     >>> from optgbm.sklearn import OGBMRegressor
@@ -892,8 +965,7 @@ class OGBMRegressor(_BaseOGBMModel, RegressorMixin):
     >>> X, y = load_boston(return_X_y=True)
     >>> reg.fit(X, y)
     OGBMRegressor(...)
-    >>> reg.score(X, y)
-    0.9...
+    >>> y_pred = reg.predict(X)
     """
 
     def __init__(
@@ -902,8 +974,8 @@ class OGBMRegressor(_BaseOGBMModel, RegressorMixin):
         num_leaves: int = 31,
         max_depth: int = -1,
         learning_rate: float = 0.1,
-        n_estimators: int = 1_000,
-        subsample_for_bin: int = 200_000,
+        n_estimators: int = 1000,
+        subsample_for_bin: int = 200000,
         objective: Optional[Union[Callable, str]] = None,
         min_split_gain: float = 0.0,
         min_child_weight: float = 1e-03,
@@ -918,14 +990,14 @@ class OGBMRegressor(_BaseOGBMModel, RegressorMixin):
         importance_type: str = "split",
         cv: Union[BaseCrossValidator, int] = 5,
         enable_pruning: bool = False,
-        n_trials: int = 10,
+        n_trials: int = 20,
         param_distributions: Optional[
             Dict[str, optuna.distributions.BaseDistribution]
         ] = None,
         refit: bool = False,
         study: Optional[optuna.study.Study] = None,
         timeout: Optional[float] = None,
-        **kwargs: Any,
+        **kwargs: Any
     ) -> None:
         super().__init__(
             boosting_type=boosting_type,
@@ -953,10 +1025,15 @@ class OGBMRegressor(_BaseOGBMModel, RegressorMixin):
             subsample_freq=subsample_freq,
             subsample_for_bin=subsample_for_bin,
             timeout=timeout,
-            **kwargs,
+            **kwargs
         )
 
-    def predict(self, X: TWO_DIM_ARRAYLIKE_TYPE) -> ONE_DIM_ARRAYLIKE_TYPE:
+    def predict(
+        self,
+        X: TWO_DIM_ARRAYLIKE_TYPE,
+        num_iteration: Optional[int] = None,
+        **predict_params: Any
+    ) -> ONE_DIM_ARRAYLIKE_TYPE:
         """Predict using the fitted model.
 
         Parameters
@@ -964,6 +1041,12 @@ class OGBMRegressor(_BaseOGBMModel, RegressorMixin):
         X
             Data.
 
+        num_iteration
+            Limit number of iterations in the prediction.
+
+        **predict_params
+            Always ignored, exists for compatibility.
+
         Returns
         -------
         y_pred
@@ -975,4 +1058,4 @@ class OGBMRegressor(_BaseOGBMModel, RegressorMixin):
             X, accept_sparse=True, estimator=self, force_all_finite=False
         )
 
-        return self._Booster.predict(X)
+        return self._Booster.predict(X, num_iteration=num_iteration)
